/*
 * parse.c - xml config file parser
 *
 * $Id$
 *
 * Copyright (c) 2001, The Colocation Corporation. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.  Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.  Neither the name of the The
 * Colocation Corporation nor the names of its contributors may be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 *
 * This software was originally written under grants from the Colocation
 * Corporation. To find the most recent versions of the software, please visit
 * http://www.loligo.com/ocoloc  To contribute patches or for other questions 
 * please contact jtodd@loligo.com
 *
 * To find out more about the Colocation Corporation, see http://www.coloco.com/ 
 *                                
 */



#define SYSLOG_NAMES 

#include <expat.h>
#include "ocoloc.h"
#include "svs.h"

#define DEBUG(fmt, va...) \
        do { if (Debug & DEBUG_PARSE) { \
             fprintf(stderr, fmt, ## va); \
             fprintf(stderr, "\n"); \
             fflush(stderr); \
        } } while (0)

XML_Parser      xmlParser;
int             DefaultsRead = 0;

enum eltTokens {
    eNotFound, eConfig, eDefaults, eDevice, eInterface,
    ePercentage, eComment, eDataPath, eCommunity, eCGIPath,
    eSysLogFac
};

char           *eltTokenNames[] = {
    "<>", "Config", "Defaults", "Device", "Interface",
    "Percentage", "Comment", "DataPath", "Community", "CGIPath", 
    "SysLogFacility", NULL
};

int
tokenToElt(const char *token)
{
    int             i = -1;
    while (eltTokenNames[++i])
	if (!(strcasecmp(token, eltTokenNames[i])))
	    return i;
    return 0;
}

#define ThisElement tokenToElt(t)

typedef struct _dataValuePair {
    int             type;
    void           *ptr;
}               DataValuePair;

typedef struct _parserUserData {
    DevList        *devtree;
    Defaults       *defaults;
    DevList        *device;
    IfList         *interface;
    int		    append_text;
}               ParserUserData;

char           *
get_attr_value(const char **list, char *key)
{
    int             i = 0;
    static char    *buf;

    if (buf)
	free(buf);

    while (list[i])
	if (!(strcasecmp(key, list[i])))
	    return (buf = strdup(list[i + 1]));

    return 0;
}

void
elt_start(void *data, const char *t, const char **attr)
{
    ParserUserData *pud = XML_GetUserData(xmlParser);
    DataValuePair  *dvp = 0;
    char           *attr_val = 0;

    if (!ThisElement)
	BAIL("Unknown XML Tag: <%s>", t);

    pud->append_text = 0;

    assert(dvp = calloc(1, sizeof(dvp)));
    dvp->type = ThisElement;

    switch (ThisElement) {

    case (eConfig):
	/* ptr = 0 from calloc */
	assert(svs_push((void *)dvp) > 0);
	DEBUG("push: %d, %p", dvp->type, dvp->ptr);
	break;

    case (eDefaults):
	dvp->ptr = pud->defaults;
	assert(svs_push((void *)dvp) > 0);
	DEBUG("push: %d, %p", dvp->type, dvp->ptr);
	break;

    case (eDevice):
	attr_val = get_attr_value(attr, "name");

	if (!pud->devtree) {
	    assert(pud->devtree = malloc(sizeof(DevList)));
	    pud->device = pud->devtree;
	} else {
	    DevList        *d = pud->device;
	    assert(pud->device = malloc(sizeof(DevList)));
	    d->next = pud->device;
	}
	memset(pud->device, 0, sizeof(pud->device));
	pud->device->name = strdup(attr_val);
	pud->interface = 0;

	dvp->ptr = pud->device;
	assert(svs_push((void *)dvp) > 0);
	DEBUG("push: %d, %p", dvp->type, dvp->ptr);
	break;

    case (eInterface):
	attr_val = get_attr_value(attr, "name");

	if (!pud->interface) {
	    assert(pud->device->interface = malloc(sizeof(IfList)));
	    pud->interface = pud->device->interface;
	} else {
	    IfList         *i = pud->interface;
	    assert(pud->interface = malloc(sizeof(IfList)));
	    i->next = pud->interface;
	}
	memset(pud->interface, 0, sizeof(pud->interface));
	pud->interface->name = strdup(attr_val);

	dvp->ptr = pud->interface;
	assert(svs_push((void *)dvp) > 0);
	DEBUG("push: %d, %p", dvp->type, dvp->ptr);
	break;

    case (eDataPath):
    case (eCommunity):
    case (ePercentage):
    case (eComment):
    case (eCGIPath):
    case (eSysLogFac):
	/* ptr repushed when we get text */
	assert(svs_push((void *)dvp) > 0);
	DEBUG("push: %d, %p", dvp->type, dvp->ptr);
	break;

    default:
	BAIL("Unknown Start Tag: <%s> in stream.", t);
    }
}

void
elt_text(void *data, const char *t, int len)
{
    ParserUserData *pud = XML_GetUserData(xmlParser);
    DataValuePair  *dvp = 0;
    int             justWhitespace = 1;
    char           *token;
    char           *ch, field[132];

    if (len == 0)
	return;

    memcpy(field, t, len);
    field[len] = '\0';
    ch = field;

    while (justWhitespace && ch < (field + len)) 
	if (*ch == ' ' || *ch == '\t' || *ch == '\n' || *ch == '\r')
	    ++ch;
	else
	    justWhitespace = 0;

    if (justWhitespace)
	return;

    assert(svs_pop((void **)&dvp) > 0);
    DEBUG(" pop: %d, %p", dvp->type, dvp->ptr);

    if (pud->append_text == 0) {
       assert(token = malloc(len + 1));
       memcpy(token, field, len + 1);
    } else {
       int oldsize = strlen((char *)dvp->ptr);
       int newsize = oldsize + len + 1;

       assert(token = malloc(newsize));
       memcpy(token, dvp->ptr, oldsize);
       memcpy(token + oldsize, field, strlen(field));
       token[newsize] = '\0';
       free(dvp->ptr);
    }

    dvp->ptr = token;
    assert(svs_push((void *)dvp) > 0);
    DEBUG("push: %d, %p", dvp->type, dvp->ptr);

    pud->append_text = 1;
}

void
elt_end(void *data, const char *t)
{
    ParserUserData *pud = XML_GetUserData(xmlParser);
    DataValuePair  *dvp, *parent;
    void           *eltValue, *parentValue;
    int             eltType, parentType;

    if (!ThisElement)
	BAIL("Unknown XML Tag: <%s>", t);

    if (ThisElement == eConfig)
	return;

    pud->append_text = 0;

    assert(svs_pop((void **)&dvp));
    DEBUG(" pop: %d, %p", dvp->type, dvp->ptr);
    eltType = dvp->type;
    eltValue = dvp->ptr;

    assert(svs_peek((void **)&parent));
    parentType = parent->type;
    parentValue = parent->ptr;

    switch (parentType) {

    case (eConfig):
	break;

    case (eDefaults):

	if (eltType == eDataPath)
	    pud->defaults->datapath = strdup((char *)eltValue);
	else if (eltType == eCommunity)
	    pud->defaults->community = strdup((char *)eltValue);
	else if (eltType == ePercentage)
	    pud->defaults->percentage = atof((char *)eltValue);
	else if (eltType == eCGIPath)
	    pud->defaults->cgipath = strdup((char *)eltValue);
	else if (eltType == eSysLogFac) {
	    int i = -1;
            while (facilitynames[++i].c_name) {
                if (!strcasecmp(facilitynames[i].c_name, (char *)eltValue)) {
	            pud->defaults->syslog_fac = facilitynames[i].c_val;
		    gvBailSysLogPri = pud->defaults->syslog_fac;
                    break;
                }
            }
            if (pud->defaults->syslog_fac == 0)
                BAIL("Invalid Syslog Facility Entered: %s", (char *)eltValue);
	} else if (eltType == eDevice);
	else
	    BAIL("Inapplicable </Defaults> sub-tag: %d on %s",
		 parentType, t);
	break;

    case (eDevice):
	if (eltType == eComment) {
	    pud->device->comment = strdup((char *)eltValue);
	    if (strchr(pud->device->comment, ':'))
		BAIL("Comments can't use ':'s, \n\t comment = %s\n",
		     pud->device->comment);
	} else if (eltType == eCommunity)
	    pud->defaults->community = strdup((char *)eltValue);
	else if (eltType == eInterface);
	else
	    BAIL("Inapplicable </Device> sub-tag: %d on %s",
		 parentType, t);
	break;

    case (eInterface):
	if (eltType == eComment) {
	    pud->interface->comment = strdup((char *)eltValue);
	    if (strchr(pud->interface->comment, ':'))
		BAIL("Comments can't use ':'s, \n\t comment = %s\n", pud->device->comment);
	} else if (eltType == ePercentage)
	    pud->interface->pct = atof((char *)eltValue);
	else
	    BAIL("Inapplicable </Interface> sub-tag: %d", parentType);
	break;

    default:
	BAIL("Unknown parentType state: %d on tag: %s",
	     parentType, t);
    }

}

int
read_conf_file(char *fn, Defaults * defaults, DevList ** hosts)
{
    FILE           *fp;
    ParserUserData  pud;
    int             doneReading = 0;

    assert(xmlParser = XML_ParserCreate(NULL));
    XML_SetElementHandler(xmlParser, elt_start, elt_end);
    XML_SetCharacterDataHandler(xmlParser, elt_text);
    XML_SetUserData(xmlParser, (void *)&pud);

    memset(&pud, 0, sizeof(pud));
    pud.defaults = defaults;

    assert(fn);
    if (!(fp = fopen(fn, "r")))
	BAIL("Couldn't open %s: %s", fn, strerror(errno));

    do {
	static char     buf[1024];
	int             len = 0;

	doneReading = feof(fp);

	len = fread(buf, sizeof(unsigned char), 1024, fp);
	if (ferror(stdin))
	    BAIL("read error on %s", fn);

	if (!XML_Parse(xmlParser, buf, len, doneReading))
	    BAIL("XML parse error at line %d of %s: %s",
		 XML_GetCurrentLineNumber(xmlParser),
		 XML_ErrorString(XML_GetErrorCode(xmlParser)),
		 fn);

    } while (!doneReading);

    *hosts = pud.devtree;

    fclose(fp);
    return 1;
}

void
dump_defaults(Defaults * d)
{
    int fac_idx = -1;

    while (facilitynames[++fac_idx].c_name)
        if (facilitynames[fac_idx].c_val == d->syslog_fac) 
            break;

    printf("\nDefaults\n");
    printf("  |-------DataPath: %s\n", d->datapath);
    printf("  |------Community: %s\n", d->community);
    printf("  |-----Percentage: %f\n", d->percentage);
    printf("  |-------CGI Path: %s\n", d->cgipath);
    printf("  |-----SysLog Fac: %s\n", facilitynames[fac_idx].c_name);
    printf("  -\n\n");
}

void
dump_device_tree(DevList * dl)
{

    DevList        *d = dl;
    IfList         *i;

    while (d) {
	printf("%s\n", (d->name ? d->name : "NULL"));

	if (d->comment)
	    printf("  |--------Comment: %s\n", d->comment);
	if (d->community)
	    printf("  |------Community: %s\n", d->community);

	i = d->interface;
	while (i) {
	    printf("  |------Interface: %s\n", (i->name ? i->name : 0));
	    if (i->comment)
		printf("           |------Comment: %s\n", i->comment);
	    if (i->pct > 0)
		printf("           |------Percentage: %0.2f\n", i->pct);
	    i = i->next;
	}

	d = d->next;
	printf("  -\n\n");
    }
}
