/*
 * poll.c - program to poll devices for statistics
 *
 * $Id$
 *
 * Copyright (c) 2001, The Colocation Corporation. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.  Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.  Neither the name of the The
 * Colocation Corporation nor the names of its contributors may be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 *
 * This software was originally written under grants from the Colocation
 * Corporation. To find the most recent versions of the software, please visit
 * http://www.loligo.com/ocoloc  To contribute patches or for other questions 
 * please contact jtodd@loligo.com
 *
 * To find out more about the Colocation Corporation, see http://www.coloco.com/ 
 *                                
 */


#include "ocoloc.h"

#define DEBUG(fmt, va...) \
        do { if (Debug & DEBUG_POLL) { \
	     fprintf(stderr, fmt, ## va); \
             fprintf(stderr, "\n"); \
	     fflush(stderr); \
	} } while (0)

extern int      Debug;

int
snmp_poll_int(char *mib, struct snmp_session * ss, long *ri)
{

   struct snmp_pdu *req, *resp;
   oid             this_oid[MAX_OID_LEN];
   size_t          this_oid_len = MAX_OID_LEN;
   int             status;

   if (!(snmp_parse_oid(mib, this_oid, &this_oid_len)))
      BAIL("snmp_parse_oid(%s) failed.", mib);

   req = snmp_pdu_create(SNMP_MSG_GET);
   snmp_add_null_var(req, this_oid, this_oid_len);

   status = snmp_synch_response(ss, req, &resp);

   if (status == STAT_SUCCESS && resp->errstat == SNMP_ERR_NOERROR) {

      if (resp->variables->type == ASN_INTEGER) {
	 *ri = *resp->variables->val.integer;
      } else
	 BAIL("snmp_get %s not returning an integer. (type=%d)",
	      mib, resp->variables->type);

   } else {

      if (status == STAT_SUCCESS)
	 fprintf(stderr, "Error in packet\nReason: %s\n",
		 snmp_errstring(resp->errstat));
      else
	 snmp_sess_perror("snmpget", ss);

      return 0;
   }

   return 1;
}

int
snmp_poll_timeticks(char *mib, struct snmp_session * ss, char **rs)
{

   struct snmp_pdu *req, *resp;
   oid             this_oid[MAX_OID_LEN];
   size_t          this_oid_len = MAX_OID_LEN;
   int             status;

   if (!(snmp_parse_oid(mib, this_oid, &this_oid_len)))
      BAIL("snmp_parse_oid(%s) failed.", mib);

   req = snmp_pdu_create(SNMP_MSG_GET);
   snmp_add_null_var(req, this_oid, this_oid_len);

   status = snmp_synch_response(ss, req, &resp);

   if (status == STAT_SUCCESS && resp->errstat == SNMP_ERR_NOERROR) {

	char tmp_buf[80], *start, *end;
            snprint_variable(tmp_buf, 79, 
		(const oid *)&this_oid, this_oid_len, resp->variables);

		start = strchr(tmp_buf, '(') + 1;
		end = strchr(tmp_buf, ')');
		*end = 0;
		*rs = strdup(start);

   } else {

      if (status == STAT_SUCCESS)
         fprintf(stderr, "Error in packet\nReason: %s\n",
                 snmp_errstring(resp->errstat));
      else
         snmp_sess_perror("snmpget", ss);

      return 0;
   }

   return 1;
}

unsigned long  *
snmp_bulk_poll_ulong(char *mib, struct snmp_session * ss, long rows)
{

   struct snmp_pdu *req, *resp;
   struct variable_list *vars;
   oid             this_oid[MAX_OID_LEN];
   size_t          this_oid_len = MAX_OID_LEN;
   unsigned long  *table;
   int             status, ti = 0;

   if (!(snmp_parse_oid(mib, this_oid, &this_oid_len)))
      BAIL("snmp_parse_oid(%s) failed.", mib);

   req = snmp_pdu_create(SNMP_MSG_GETBULK);
   req->non_repeaters = 0;
   req->max_repetitions = rows;
   snmp_add_null_var(req, this_oid, this_oid_len);

   status = snmp_synch_response(ss, req, &resp);

   if (status == STAT_SUCCESS && resp->errstat == SNMP_ERR_NOERROR) {

      assert(table = calloc(rows, sizeof(unsigned long)));

      for (vars = resp->variables; vars; vars = vars->next_variable) {

	 if (vars->type == ASN_COUNTER) {
	    table[ti] = (unsigned long)*vars->val.integer;

	 } else
	    BAIL("snmp_get %s not an integer or string. (type=%d)",
		 mib, vars->type);

	 ti += 1;
      }

   } else {

      if (status == STAT_SUCCESS)
	 fprintf(stderr, "Error in packet\nReason: %s\n",
		 snmp_errstring(resp->errstat));
      else
	 snmp_sess_perror("snmpget", ss);

      return 0;
   }

#if 0
   for (ti = 0; ti < rows; ti++)
      printf("[%d] ul = %lu\n", ti, table[ti]);
   printf("\n");
#endif

   return table;
}

char          **
snmp_bulk_poll_char(char *mib, struct snmp_session * ss, long rows)
{
   struct snmp_pdu *req, *resp;
   struct variable_list *vars;
   oid             this_oid[MAX_OID_LEN];
   size_t          this_oid_len = MAX_OID_LEN;
   char          **table;
   int             status, ti = 0;

   if (!(snmp_parse_oid(mib, this_oid, &this_oid_len)))
      BAIL("snmp_parse_oid(%s) failed.", mib);

   req = snmp_pdu_create(SNMP_MSG_GETBULK);
   req->non_repeaters = 0;
   req->max_repetitions = rows;
   snmp_add_null_var(req, this_oid, this_oid_len);

   status = snmp_synch_response(ss, req, &resp);

   assert(table = calloc(rows, sizeof(char *)));

   if (status == STAT_SUCCESS && resp->errstat == SNMP_ERR_NOERROR) {

      for (vars = resp->variables; vars; vars = vars->next_variable) {

	 if (vars->type == ASN_OCTET_STR) {
	    assert(table[ti] = calloc(1, vars->val_len + 1));
	    memcpy(table[ti], vars->val.string, vars->val_len);
	 } else
	    BAIL("snmp_get %s not an integer or string. (type=%d)",
		 mib, vars->type);

	 ti += 1;
      }

   } else {

      if (status == STAT_SUCCESS)
	 fprintf(stderr, "Error in packet\nReason: %s\n",
		 snmp_errstring(resp->errstat));
      else
	 snmp_sess_perror("snmpget", ss);

      return 0;
   }

   return table;
}

void
poll_device(DevList * host, char *def_community)
{
   char           *mib_ifCount = "interfaces.ifNumber.0";
   char           *mib_sysUpTime = "system.sysUpTime.0";
   char           *mib_ifDesc = "interfaces.ifTable.ifEntry.ifDescr";
   char           *mib_ifIn = "interfaces.ifTable.ifEntry.ifInOctets";
   char           *mib_ifOut = "interfaces.ifTable.ifEntry.ifOutOctets";

   struct snmp_session session, *ss;
   IfList         *ifc = host->interface;
   long            if_count = -1;
   char          **if_desc, *sys_uptime;
   unsigned long  *if_in;
   unsigned long  *if_out;

   time_t          inTime, outTime;

   if (!host)
      BAIL("poll_device called with NULL host");

   DEBUG("\nPolling Host: %s", host->name);

   if (!host->community && !def_community)
      BAIL("poll_device called and no community info provided.");

   snmp_sess_init(&session);
   session.version = SNMP_VERSION_2c;
   session.peername = strdup(host->name);
   session.community = strdup(host->community ? host->community : def_community);
   session.community_len = strlen(session.community);

   if (!(ss = snmp_open(&session))) {
      snmp_perror("snmp_open");
      snmp_sess_perror("session: snmp_open", &session);
      exit(1);
   }
   if (!(snmp_poll_int(mib_ifCount, ss, &if_count)))
      BAIL("Interface count on %s failed.", mib_ifCount);

   DEBUG("\tInterface Count: %ld", if_count);

   if (!(if_desc = snmp_bulk_poll_char(mib_ifDesc, ss, if_count)))
      BAIL("Description bulk poll on %s failed.", mib_ifDesc);

   while (ifc) {
      int             i;
      for (i = 0; i < if_count && strcasecmp(ifc->name, if_desc[i]); i++);

      if (i == if_count)
	 BAIL("Unknown Port name: %s", ifc->name);

      ifc->index = i;
      DEBUG("\t%s is ifDesc[%d]", ifc->name, i);
      ifc = ifc->next;
   }

   /* XXX dispose of table properly later, when time isn't important XXX */

   if (!(if_in = snmp_bulk_poll_ulong(mib_ifIn, ss, if_count)))
      BAIL("Input bulk poll on %s failed.", mib_ifIn);
   inTime = time((time_t *) NULL);

   if (!(if_out = snmp_bulk_poll_ulong(mib_ifOut, ss, if_count)))
      BAIL("Output bulk poll on %s failed.", mib_ifOut);
   outTime = time((time_t *) NULL);

   if (!(snmp_poll_timeticks(mib_sysUpTime, ss, &sys_uptime)))
      BAIL("Output poll on %s failed.", mib_sysUpTime);

   ifc = host->interface;
   while (ifc) {

      ifc->inOctets = if_in[ifc->index];
      ifc->inTime = inTime;

      ifc->outOctets = if_out[ifc->index];
      ifc->outTime = outTime;

      ifc->upTime = sys_uptime;

      DEBUG("\t%s in: %lu, out: %lu up: %s",
	    ifc->name, ifc->inOctets, ifc->outOctets, ifc->upTime);
      ifc = ifc->next;
   }

}

void
write_samples(Defaults * def, DevList * host)
{
   IfList         *ifc = host->interface;
   char            dir[128], fn[128];
   char           *ch;

   DEBUG("\nWriting Samples for Host: %s", host->name);

   sprintf(dir, "%s/%s", def->datapath, host->name);
   if (mkdir(dir, 0755) == -1) {
      if (errno != EEXIST)
	 BAIL("Can't mkdir(%s)", dir);
   } else
      DEBUG("Created %s", dir);

   while (ifc) {

      ch = ifc->name;
      while ((ch = strchr(ch, '/')))
	 *ch = '_';

      sprintf(fn, "%s/%s/%s", def->datapath, host->name, ifc->name);

      if (!(ifc->fp = fopen(fn, "a"))) {
	 /* log couldn't open file */
	 perror("fopen()");
	 continue;
      }
      fprintf(ifc->fp, "%s %u %lu %u %lu\n",
	      ifc->upTime, ifc->inTime, ifc->inOctets, ifc->outTime, 
	ifc->outOctets);

      DEBUG("%20s %u %lu %u %lu", ifc->name,
	    ifc->inTime, ifc->inOctets, ifc->outTime, ifc->outOctets);

      fclose(ifc->fp);
      ifc = ifc->next;
   }
}


int
poll_device_tree(Options *opt, Defaults * def, DevList * hosttree)
{
   DevList        *host;
   pid_t           pid, ppid = getpid();
   int             children = 0, waitrv = 0;

   init_snmp("ocoloc");

   for (host = hosttree; host; host = host->next) {

      if (Debug || opt->dont_fork) {
	 poll_device(host, def->community);
	 write_samples(def, host);
      } else {

	 if (!(pid = fork())) {
	    setproctitle("%s [%d]", host->name, ppid);
	    poll_device(host, def->community);
	    write_samples(def, host);
	    exit(0);
	 }
	 /* Parent fork */
	 children++;
	 DEBUG("fork()ing for %s, pid = %d, %d children\n",
	       host->name, pid, children);
	 setproctitle("ocoloc parent");
      }
   }

   while (children) {
      DEBUG("waiting on %d children... ", children);

      wait(&waitrv);
      DEBUG("zombie: %d\n", waitrv);
      --children;
   }

   DEBUG("done.");
   return 1;
}
